import sys
import os
import time
import threading
import logging
import fnmatch

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, FileDeletedEvent

__all__ = [
    'setup_watch',
]

logger = logging.getLogger('jinjastatic')


class EventHandler(FileSystemEventHandler):
    def __init__(self, base_path, callback, excludes=[]):
        self.base_path = os.path.abspath(base_path)
        self.modified_files = set()
        self.promise = None
        self.callback = callback
        self.excludes = excludes

    def on_any_event(self, event):
        if event.is_directory:
            return
        path = event.src_path
        path = path[len(self.base_path):].lstrip('/')
        for exclude in self.excludes:
            if fnmatch.fnmatch(path, exclude):
                return

        if isinstance(event, FileDeletedEvent):
            if path in self.modified_files:
                self.modified_files.remove(path)
            return

        self.modified_files.add(path)
        if not self.promise or self.promise.cancel():
            self.promise = DelayedPromise(self.run_combined, 0.1)

    def run_combined(self):
        f = self.modified_files
        self.modified_files = set()
        self.callback(list(f))


def setup_watch(src_dir, callback, excludes=[]):
    observer = Observer()
    observer.schedule(EventHandler(src_dir, callback, excludes), path=src_dir, recursive=True)
    observer.start()
    logger.info("Set up watching on {0} recursively. (CTRL-C to stop)".format(src_dir))
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


class DelayedPromise(threading.Thread):
    def __init__(self, callback, delay_time=0.1):
        threading.Thread.__init__(self)
        self.cancelled = False
        self.running = threading.Lock()
        self.delay_time = delay_time
        self.stopped = False
        self.callback = callback
        self.setDaemon(True)
        self.start()


    def cancel(self):
        if self.running.acquire(False) or self.stopped:
            self.cancelled = True
            self.running.release()
            return True
        return False

    def run(self):
        time.sleep(self.delay_time)
        self.running.acquire()
        if self.cancelled:
            return
        self.stopped = True
        try:
            self.callback()
        finally:
            self.running.release()



import os
import sys
import argparse
import logging
import shutil

import yaml
import jinja2
import logging
import jinjatag
import datetime


from utils import is_updated
import staticlib
from watcher import setup_watch
from dependencies import Dependencies

logger = logging.getLogger('jinjastatic')

def run():
    configure_logging()

    p = argparse.ArgumentParser(description="Compile static templates")
    p.add_argument('-s', '--source', required=True,
                   help="Source file or directory.")
    p.add_argument('-d', '--dest',
                   help="Destination file or directory.")
    p.add_argument('-w', '--watch', action="store_true", default=False,
                   help="Watch for changed files.")
    p.add_argument('-f', '--full', action="store_true", default=False,
                   help="Do not perform an incremental compilation, and do everything.")
    p.add_argument('-p', '--production', action="store_true", default=False,
                   help="Minify and compile static files for use in production.")
    p.add_argument('-c', '--config', default='config.yml',
                   help='Name of config file with settings.')
    p.add_argument('-q', '--quiet', action="store_true", default=False,
                   help='Suppress output')
    p.add_argument('-x', '--compiledir', default=None,
                   help='Default directory to house compiled files.')
    args = p.parse_args()

    if args.quiet:
        logger.setLevel(logging.ERROR)

    if args.production:
        if args.compiledir:
            compiledir = args.compiledir
        else:
            compiledir = os.path.join(args.dest, 'compiled')


        if not os.path.exists(compiledir):
            os.makedirs(compiledir)
    else:
        compiledir = None

    config = {}
    if os.path.exists(args.config):
        with open(args.config) as f:
            config = yaml.load(f.read())

    env, loader = get_jinja_env(args.source)
    dependencies = Dependencies(args.source, env, loader)
    dependencies.load_graph()

    if args.watch:
        compile_jinja(args.source, args.dest, config, True, True, compiledir, dependencies)
        setup_watch(args.source,
                    FileHandler(args.source, args.dest, config, dependencies),
                    ['.*', '*#*', '*~'],
                    )
        return

    compile_jinja(args.source, args.dest, config, not args.full and not args.production, not args.production, compiledir, dependencies)


def compile_jinja(source, dest, config, incremental, debug, compiledir, dependencies):
    env = get_jinja_env(source)[0]

    staticlib.clear_data()

    staticlib.set_config(debug, config, source)

    if incremental:
        changed = walk_for_changed(source, dest, dependencies)
    else:
        changed = ()

    if not debug:
        walk_and_compile(env, source, dest, incremental, False, changed)
        staticlib.compile(source, compiledir, dest)
    walk_and_compile(env, source, dest, incremental, True, changed)

def walk_and_compile(env, source, dest, incremental, save, changed):
    for dirpath, dirnames, filenames in os.walk(source, followlinks=True):
        reldir = dirpath[len(source):].lstrip('/')
        for filename in filenames:
            if not filename.lower().endswith('.html'):
                if save:
                    copy_file(os.path.join(source, reldir, filename),
                              os.path.join(dest, reldir, filename),
                              incremental)
                continue
            if save:
                target_file = os.path.join(dest, reldir, filename)
            else:
                target_file = None
            try:
                if not incremental or os.path.join(reldir, filename) in changed:
                    compile_file(env, os.path.join(reldir, filename),
                                 os.path.join(dirpath, filename), target_file, False)
            except Exception as e:
                logger.error("   In file {0}: {1}".format(os.path.join(reldir, filename),
                                                        str(e)), exc_info=True)

def walk_for_changed(source, dest, dependencies):
    changed = set()
    for dirpath, dirnames, filenames in os.walk(source, followlinks=True):
        reldir = dirpath[len(source):].lstrip('/')
        for filename in filenames:
            if not filename.lower().endswith('.html'):
                continue
            name = os.path.join(reldir, filename)
            if is_updated(os.path.join(source, reldir, filename),
                          os.path.join(dest, reldir, filename)):
                changed.add(name)
                changed.update(dependencies.get_affected_files(name))
    return changed


class FileHandler(object):
    def __init__(self, source, dest, config, dependencies):
        self.source = source
        self.dest = dest
        self.config = config
        self.dependencies = dependencies

    def __call__(self, files):
        env = get_jinja_env(self.source)[0]
        for fname in files:
            self.dependencies.recompute_file(fname)
        total_changed = set()
        for fname in files:
            total_changed.add(fname)
            total_changed.update(self.dependencies.get_affected_files(fname))
        for fname in total_changed:
            if fname.lower().endswith('.html'):
                target_file = os.path.join(self.dest, fname)
                compile_file(env, fname, os.path.join(self.source, fname),
                             target_file, False)
            else:
                copy_file(os.path.join(self.source, fname),
                          os.path.join(self.dest, fname),
                          True)

def copy_file(source, dest, incremental):
    if not incremental or is_updated(source, dest):
        if not os.path.exists(os.path.dirname(dest)):
            os.makedirs(os.path.dirname(dest))
        shutil.copyfile(source, dest)

def compile_file(env, source_name, source_file, dest_file, incremental):
    if incremental and not is_updated(source_file, dest_file):
        return

    if dest_file:
        logger.debug("Compiling {0} -> {1}".format(source_file, dest_file))
    ctx = {
        'datetime': datetime,
        'env': EnvWrapper(),
        }
    try:
        result = env.get_template(source_name).render(ctx).encode('utf8')
    except Exception as e:
        logger.error("Error compiling {0}".format(source_name), exc_info=True)
        return
    if not dest_file:
        return
    with with_dir(open, dest_file, 'w+') as f:
        f.write(result)

def get_jinja_env(source):
    jinja_tag = jinjatag.JinjaTag()
    loader = jinja2.FileSystemLoader(source)
    env = jinja2.Environment(loader=loader, extensions=[jinja_tag])
    jinja_tag.init()
    return env, loader

def with_dir(callback, filename, *args, **kwargs):
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    return callback(filename, *args, **kwargs)

def configure_logging():
    logger.setLevel(logging.DEBUG)
    fh = logging.StreamHandler()
    formatter = logging.Formatter("%(message)s")
    fh.setFormatter(formatter)
    logger.addHandler(fh)


class EnvWrapper(object):
    def __getattr__(self, name):
        return os.environ.get(name)



if __name__ == '__main__':
    run()



import os
import glob
import time
import pipes
import shutil
import random
import tempfile
import hashlib
import logging
import collections
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import envoy
import jinjatagext

from utils import is_updated

logger = logging.getLogger('jinjastatic')

extensions = {
    'text/css': 'css',
    'text/javascript': 'js',
    }

pre_compilers = {
    'text/less': ('lessc %(input)s', 'text/css', 'css',),
    'text/coffeescript': ('coffee --join %(output)s -c %(input)s', 'text/javascript', 'js',),
    }

compilers = {
    'text/javascript': './runyui.sh --nomunge %(input)s',
    'text/css': 'uglifycss %(input)s',
    }


g = {
    'debug': False,
    ('text/css', False): {},
    ('text/javascript', False): {},
    ('text/css', True): {},
    ('text/javascript', True): {},
    'temp_dir': tempfile.mkdtemp(prefix='jinjastatic'),
    'config': {},
    'compiled': {},
    'compsheets': {},
    'minified': {},
}

def _handle_tag(type_, ctx, src, debug=False, head=False, **kwargs):
    kwargs.setdefault('type', type_)
    format = style_formats[type_]
    if g['debug']:
        return format.format(
            src,
            u' '.join('{0}={1!r}'.format(k, _force_str(v)) for k, v in kwargs.items()))
    elif debug:
        return u''
    key = (type_, head)
    compiled_key = (pre_compilers.get(type_, (None, type_))[1], head)
    format = style_formats[compiled_key[0]]
    ctxname = ctx.name
    min_dict = g['minified'].setdefault(ctxname, {})

    if min_dict.get(compiled_key):
        return ''
    elif ctxname in g.get(compiled_key, {}) and g['compiled']:
        files = OrderedDict([(g['compiled'][orig], 1)
                             for orig in g[compiled_key][ctxname]]).keys()
        min_dict[compiled_key] = True
        return u'\n'.join(format.format(src, 'type="{0}"'.format(compiled_key[0]))
                          for src in files)
    elif type_ in pre_compilers:
        pre_compile(src, type_, head, ctxname)
        return u''
    else:
        g[key].setdefault(ctxname, []).append(src.lstrip('/'))
    return '**FIRSTPASS**'

style_formats = {
    'text/javascript': u'<script src="{0}" {1}></script>',
    'text/css': u'<link rel="stylesheet" href="{0}" {1}>',
    'text/less': u'<link rel="stylesheet/less" href="{0}" {1}>',
    'text/coffeescript': u'<script src="{0}" {1}></script>',
}

@jinjatagext.simple_context_tag
def script(ctx, src, **kwargs):
    return _handle_tag(u'text/javascript', ctx,
                       src, **kwargs)


@jinjatagext.simple_context_tag
def style(ctx, href, **kwargs):
    return _handle_tag(u'text/css', ctx, href, **kwargs)

@jinjatagext.simple_context_tag
def less(ctx, href, **kwargs):
    return _handle_tag(u'text/less', ctx, href, **kwargs)

@jinjatagext.simple_context_tag
def coffee(ctx, src, **kwargs):
    return _handle_tag(u'text/coffeescript', ctx,
                       src, **kwargs)

def clear_data():
    try:
        shutil.rmtree(g['temp_dir'])
    except:
        pass
    g.update({
            ('text/css', False): {},
            ('text/javascript', False): {},
            ('text/css', True): {},
            ('text/javascript', True): {},
            'compiled': {},
            'temp_dir': tempfile.mkdtemp(prefix='jinjastatic'),
            'minified': {},
            })

def set_config(debug, config, base_dir):
    g['debug'] = debug
    g['base_dir'] = base_dir
    mapper = config.get('map', {})
    config['map'] = {}
    for k, v in mapper.items():
        for f in v:
            config['map'][f] = k
    g['config'] = config

def compile(base_dir, output_dir, dest_dir):
    config = g['config']
    _remove_old_files(output_dir)

    rel_output = '/' + output_dir[len(dest_dir):].lstrip('/')

    static_dirs = set()

    for key in g:
        if not isinstance(key, tuple):
            continue
        unique_key = _gen_key()
        filemap = g[key]
        ext = extensions[key[0]]
        files = OrderedDict((filename, 1) for template in filemap for filename in filemap[template]).keys()
        file_comp = collections.defaultdict(list)
        for filename in files:
            aggregate = _decorate_key(config['map'].get(filename, 'maincompiled.' + ext), unique_key)
            file_comp[aggregate].append(filename)
        for target, filelist in file_comp.items():
            abstarget = os.path.join(output_dir, target)
            absfilelist = [os.path.join(base_dir, filename) for filename in filelist]
            logger.info('Compiling {0}'.format(abstarget))
            combined_file_obj = None
            if len(absfilelist) == 1:
                combined_file = absfilelist[0]
            else:
                combined_file_obj = _combine_files(absfilelist, ext)
                combined_file = combined_file_obj.name

            compiler_fmt = compilers[key[0]]
            if '%(input)s' in compiler_fmt:
                data = None
                cmd = compiler_fmt % {'input': combined_file}
            else:
                data = read_file_data([combined_file])
                cmd = compiler_fmt
            output = envoy.run(cmd, data=data)
            if output.status_code:
                raise RuntimeError(output.std_err)
            with open(abstarget, 'wb+') as f:
                f.write(output.std_out)
            target = os.path.join(rel_output, target)
            g['compiled'].update(dict((filename, target) for filename in filelist))

            if key[0] == 'text/css':
                static_dirs.update(os.path.dirname(filename) for filename in absfilelist)

    for d in static_dirs:
        for dirpath, dirnames, filenames in os.walk(d, followlinks=True):
            reldir = dirpath[len(d):].lstrip('/')
            target_dir = os.path.join(output_dir, reldir)
            for filepath in filenames:
                target_path = os.path.join(target_dir, filepath)
                filepath = os.path.join(dirpath, filepath)
                if is_updated(filepath, target_path):
                    new_dir = os.path.dirname(target_path)
                    if not os.path.exists(new_dir):
                        os.makedirs(new_dir)
                    shutil.copyfile(filepath, target_path)

g = {
    'debug': False,
    ('text/css', False): {},
    ('text/javascript', False): {},
    ('text/css', True): {},
    ('text/javascript', True): {},
    'temp_dir': tempfile.mkdtemp(prefix='jinjastatic'),
    'config': {},
    'compiled': {},
    'compsheets': {},
    'minified': {},
}

def pre_compile(src, type_, head, ctxname):
    compiler, type_, ext = pre_compilers[type_]
    key = (type_, head)
    script_list = g[key].setdefault(ctxname, [])
    old_file = os.path.join(g['base_dir'], src.lstrip('/'))
    new_name = os.path.join(os.path.dirname(src).strip('/'), 'compiled-' + hashlib.md5(src).hexdigest()) + "." + ext

    if new_name in script_list:
        return

    script_list.append(new_name)

    new_name = os.path.join(g['base_dir'], new_name.lstrip('/'))

    if not is_updated(old_file, new_name):
        return

    logger.info('Pre-compiling {0} -> {1}'.format(old_file, new_name))

    params = {'input': pipes.quote(old_file)}
    use_stdout = True

    if '%(output)s' in compiler:
        use_stdout = False
        params['output'] = pipes.quote(new_name)

    output = envoy.run(compiler % params)
    if output.status_code:
        raise RuntimeError(output.std_err)

    if not use_stdout:
        return
    with open(new_name, 'wb+') as f:
        f.write(output.std_out)

def read_file_data(filelist):
    result = []
    for fname in filelist:
        with open(fname) as f:
            result.append(f.read())
    return ''.join(result)

def _remove_old_files(output_dir):
    for fname in glob.glob(os.path.join(output_dir, '*_min*')):
        os.unlink(fname)

def _combine_files(filelist, ext):
    tmp = tempfile.NamedTemporaryFile(suffix='.' + ext)
    for filename in filelist:
        with open(filename, 'rb') as f:
            shutil.copyfileobj(f, tmp)
        if ext.lower() == 'js':
            tmp.write(';')
    tmp.flush()
    return tmp

def _decorate_key(filename, key):
    first, second = filename.rsplit('.', 1)
    return first + '-' + key + '.' + second

def _gen_key():
    return hashlib.md5(str(time.time() + random.random())).hexdigest()[:10] + '_min'


def _force_str(obj):
    if isinstance(obj, unicode):
        return obj.encode('utf8')
    return str(obj)



import os
import jinja2
from jinja2 import meta
import networkx as nx
import logging

logger = logging.getLogger('jinjastatic')

class Dependencies(object):
    def __init__(self, source, env, loader):
        self.env = env
        self.loader = loader
        self.source = source
        self.dependency_graph = nx.DiGraph()

    def get_affected_files(self, template, acc=None):
        if not acc:
            acc = set()
        if template in self.dependency_graph:
            for child in self.dependency_graph.successors(template):
                acc.add(child)
                self.get_affected_files(child, acc)
        return list(acc)

    def load_graph(self):
        self.depencency_graph = nx.DiGraph()
        for relpath, dirnames, filenames in os.walk(self.source):
            for filename in filenames:
                if not filename.endswith('.html'):
                    continue
                path = os.path.join(relpath, filename)
                name = path[len(self.source):].lstrip('/')
                requirements = self._get_requirements(name)
                for requirement in requirements:
                    if requirement:
                        self.dependency_graph.add_edge(requirement, name)

    def recompute_file(self, template):
        old_attached = []
        if not template.endswith('.html'):
            return
        if template in self.dependency_graph:
            old_attached = list(self.dependency_graph.successors(template))
            self.dependency_graph.remove_node(template)
        for requirement in self._get_requirements(template):
            if requirement:
                self.dependency_graph.add_edge(requirement, template)
        for old_name in old_attached:
            self.dependency_graph.add_edge(template, old_name)

    def _get_requirements(self, template_name):
        try:
            return list(meta.find_referenced_templates(self.env.parse(self.loader.get_source(self.env, template_name))))
        except Exception as e:
            logger.error("Error analyzing {0}".format(template_name), exc_info=True)
        return []




# -*- coding: utf-8 -
#


import os
from setuptools import setup, find_packages
import sys


requirements = [
    'networkx',
    'jinja2',
    'jinjatag',
    'watchdog==0.6.0',
    'pyyaml',
    'envoy==0.0.2',
    ]

if sys.version_info[0] == 2 and sys.version_info[1] < 7:
    requirements.append('argparse')
    requirements.append('ordereddict')

setup(
    name = 'jinjastatic',
    version = '0.0.3',

    description = 'Static template compilation',
    long_description = '',
    author = 'Michael Axiak',
    author_email = 'mike@axiak.net',
    license = 'MIT',

    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    zip_safe = False,
    packages = find_packages(exclude=['examples', 'tests']),
    include_package_data = True,
    install_requires=requirements,
    entry_points={
        'console_scripts':
            ['jinja-static=jinjastatic:run'],
        },
)



from jinja2 import nodes
try:
    from jinjatag import simple_context_tag
except ImportError:
    from jinjatag import decorators
    @decorators.create_extension_decorator
    class simple_context_tag(decorators.BaseTag):
        def parse_attrs(self, parser, add_id=True, with_context=False):
            attrs = {}
            while parser.stream.current.type != 'block_end':
                node = parser.parse_assign_target(with_tuple=False)

                if parser.stream.skip_if('assign'):
                    attrs[node.name] = parser.parse_expression()
                else:
                    attrs[node.name] = nodes.Const(node.name)
            if with_context:
                attrs['ctx'] = nodes.ContextReference()
            return nodes.Dict([nodes.Pair(nodes.Const(k), v) for k,v in attrs.items()])

        def parse(self, parser):
            tag = parser.stream.next()
            attrs = self.parse_attrs(parser, with_context=True)
            return nodes.Output([self.call_method('_call_simple_tag', args=[attrs])])

        def _call_simple_tag(self, attrs):
            return self.call_tag_func(**attrs)
